import importlib
import os
import tomllib
from collections.abc import Generator, Mapping
from dataclasses import dataclass
from importlib.resources.abc import Traversable
from pathlib import Path
from typing import Any

import unistrant.data
from unistrant.error import UnistrantError


@dataclass(frozen=True)
class Settings(Mapping):
    settings: Mapping[str, Any]
    logging: Mapping[str, Any]

    def __contains__(self, key: Any) -> bool:
        return key in self.settings

    def __getitem__(self, key: Any) -> Any:
        return self.settings[key]

    def __iter__(self) -> Generator:
        yield from self.settings

    def __len__(self) -> int:
        return len(self.settings)

    def required(self, key: Any) -> Any:
        try:
            return self.settings[key]
        except KeyError:
            raise UnistrantError(f"Required setting {key} missing")


def find_settings(settings_path: Path | None) -> Path | None:
    if settings_path:
        return settings_path

    if home := os.getenv("HOME"):
        home_settings_path = (Path(home) / ".config/sams/unistrant/unistrant.toml").expanduser()
        if home_settings_path.exists():
            return home_settings_path

    system_settings_path = Path("/etc/sams/unistrant/unistrant.toml")
    if system_settings_path.exists():
        return system_settings_path

    return None


def load_settings(settings_path: Path | None) -> Settings:
    settings = {}

    def merge(path: Traversable) -> None:
        try:
            with path.open(mode="rb") as settings_file:
                toml = tomllib.load(settings_file)
        except OSError as e:
            raise UnistrantError(f"Unable to read settings file: {str(e)}")
        except tomllib.TOMLDecodeError as e:
            raise UnistrantError(f"Unable to decode settings file: {str(e)}")
        settings.update(toml)

    merge(importlib.resources.files(unistrant.data) / "settings.toml")

    if settings_path is not None:
        merge(settings_path)

    def flatten(s: dict, base: str = "") -> Any:
        for key, value in s.items():
            full_key = f"{base}.{key}" if base else key
            if isinstance(value, dict):
                yield from flatten(value, full_key)
            else:
                yield full_key, value

    return Settings(settings=dict(flatten(settings)), logging=settings.get("logging", {}))



import datetime
import logging
import shutil
from abc import ABC, abstractmethod
from collections.abc import Set
from pathlib import Path

from unistrant.http import CertificateAuthentication, HttpProtocol
from unistrant.options import Options
from unistrant.record import RecordDocument
from unistrant.sams import SamsClient

logger = logging.getLogger(__name__)


class BaseCommand(ABC):
    def __init__(self, options: Options):
        super().__init__()
        self.options = options
        self.error = False

        authentication = CertificateAuthentication(options.sams_certificate, options.sams_key)
        protocol = HttpProtocol(authentication)
        self.sams = SamsClient(options.sams_url, protocol)

    @abstractmethod
    def run(self) -> None:
        pass

    def fail(self) -> None:
        if not self.error:
            self.error = True

    @property
    def record_files(self) -> Set[Path]:
        return {file for file in self.options.records_directory.iterdir() if file.is_file()}


class RegisterCommand(BaseCommand):
    def run(self) -> None:
        for path in self.record_files:
            logger.info(f"Processing {path.name}")
            try:
                self.process(path)
            except Exception as e:
                logger.error(f"Error processing {path.name}: {str(e)}")
                self.fail()

    def process(self, path: Path) -> None:
        self.upload(path)
        self.archive(path)

    def upload(self, path: Path) -> None:
        document = RecordDocument(path)
        self.sams.upload_record_document(document)

    def archive(self, path: Path) -> None:
        if (destination := self.options.archive_directory / path.name).exists():
            timestamp = datetime.datetime.now(datetime.UTC).astimezone().isoformat()
            destination = destination.with_name(f"{path.name}-{timestamp}")
        logger.debug(f"Archiving {path.name} to {destination}")
        shutil.move(path, destination)



import enum
import xml.etree.ElementTree as ET
from pathlib import Path

from unistrant.error import RecordError


class RecordType(enum.StrEnum):
    UsageRecord = "{http://schema.ogf.org/urf/2003/09/urf}UsageRecords"
    StorageUsageRecord = "{http://eu-emi.eu/namespaces/2011/02/storagerecord}StorageUsageRecords"
    CloudRecord = "{http://sams.snic.se/namespaces/2016/04/cloudrecords}CloudRecords"
    SoftwareAccountingRecord = "{http://sams.snic.se/namespaces/2019/01/softwareaccountingrecords}SoftwareAccountingRecord"


class RecordDocument:
    def __init__(self, path: Path):
        with path.open("rb") as f:
            try:
                tree = ET.parse(f)
            except ET.ParseError as e:
                raise RecordError(f"Record parser error: {str(e)}")
        ET.indent(tree)
        if root := tree.getroot():
            root_tag = root.tag
        else:
            raise RecordError("Missing root tag")
        try:
            self._record_file_type = RecordType(root_tag)
        except ValueError:
            raise RecordError(f"Unsupported root tag: {root_tag}")
        self.tree = tree

    @property
    def record_file_type(self) -> RecordType:
        return self._record_file_type

    @property
    def bytes(self) -> bytes:
        return ET.tostring(self.tree.getroot())
